/**
  Generated Main Source File

  Company:
    Microchip Technology Inc.

  File Name:
    main.c

  Summary:
    This is the main file generated using MPLAB® Code Configurator

  Description:
    This header file provides implementations for driver APIs for all modules selected in the GUI.
    Generation Information :
        Product Revision  :  MPLAB® Code Configurator - v2.25.2
        Device            :  PIC16F1827
        Driver Version    :  2.00
    The generated drivers are tested against the following:
        Compiler          :  XC8 v1.34
        MPLAB             :  MPLAB X v2.35 or v3.00
 */

/*
Copyright (c) 2013 - 2015 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 */

#include "mcc_generated_files/mcc.h"

#define U_MUL   19              // U analog input to mV ratio
#define I_MUL   30              // I analog input to mA ratio

#define U_MIN 10000             // 10 V - min battery voltage 
#define U_MAX 15000             // 15 V - max battery voltage
#define I_MAX 10000             // 10 A - max current

#define U_DISCHARGED 12000      // 12 V - bulk charge treshold
#define U_CHARGED 14400         // 14.4 V - absorb mode voltage / treshold
#define U_MAINTAIN 13650        // 13.65 V - maintain mode voltage
#define I_PRECHARGE 150         // 150 mA - precharge current limit
#define I_CHARGE 8000           // 8 A - bulk charge current limit
#define I_MAINTAIN 1000         // 1 A - maintain mode current treshold 

#define PWR_MAX_DUTY 500        // max. PWM duty value (the trasformer based driver cannot work with 100% duty)

#define G_FULL_DUTY  405        // galvanometer full scale PWM value (9bit resolution)
#define G_HIGH_SCALE_MIN 0
#define G_HIGH_SCALE_MAX 10000  // milli
#define G_LOW_SCALE_MIN 10000   // milli
#define G_LOW_SCALE_MAX 15000   // milli

#define PEEK_FILTER_Q 4         // analog input SW filter coefficient

#define LEAD_CONDUCTANCE 17     // 1/0.06 Ohm cable conductance between source and accumulator, voltage compensation

enum {
    error,
    pre_charge,
    bulk_charge,
    absorb_charge,
    maintain_charge
} mode;

static uint16_t Out_duty;

static uint16_t U_bat; // mV
static uint16_t I_bat; // mA

void increase_output() {
    if (Out_duty < PWR_MAX_DUTY)
        Out_duty++;
    EPWM1_LoadDutyValue(Out_duty);
}

void decrease_output() {
    if (Out_duty > 0)
        Out_duty--;
    EPWM1_LoadDutyValue(Out_duty);
}

void set_output(uint16_t u_limit, uint16_t i_limit) {
    if (I_bat < i_limit && U_bat < u_limit)
        increase_output();

    if (I_bat > i_limit || U_bat > u_limit)
        decrease_output();
}

void show_value(uint16_t value, uint16_t scale_min, uint16_t scale_max) {
    if (value < scale_min)
        EPWM2_LoadDutyValue(0);
    else {
        uint32_t result = (uint32_t)(value - scale_min) * G_FULL_DUTY;
        EPWM2_LoadDutyValue(result / (scale_max - scale_min));
    }
}

/** works with 10bit values from AD converter, so uint16 is large enough for PEEK_FILTER_Q up to 63 */
uint16_t Filter(uint16_t average, uint16_t current) {
    average *= PEEK_FILTER_Q - 1;
    average += current;
    return average / PEEK_FILTER_Q;
}

void main(void) {

    uint16_t ADC_high = 0;
    uint16_t ADC_low = 0;

    // initialize the device
    SYSTEM_Initialize();

    Out_duty = 0;
    EPWM1_LoadDutyValue(0);
    EPWM2_LoadDutyValue(0);

    INTERRUPT_GlobalInterruptEnable();
    INTERRUPT_PeripheralInterruptEnable();

    mode = error;
    RELAY_LAT = 0;

    while (1) {

        CLRWDT();

        ADC_high = Filter(ADC_high, ADC_GetConversion(U_sense));
        ADC_low = Filter(ADC_low, ADC_GetConversion(I_sense));

        U_bat = (ADC_high * U_MUL) - ADC_low;
        I_bat = ADC_low * I_MUL;
        U_bat -= I_bat / LEAD_CONDUCTANCE;

        switch (mode) {

            case pre_charge:
                set_output(U_CHARGED, I_PRECHARGE);
                break;

            case bulk_charge:
            case absorb_charge:
                set_output(U_CHARGED, I_CHARGE);
                break;

            case maintain_charge:
                set_output(U_MAINTAIN, I_CHARGE);
                break;
        }

        if (U_I_SWITCH_PORT) {
            if (mode != error) {
                show_value(U_bat, G_LOW_SCALE_MIN, G_LOW_SCALE_MAX);
            } else {
                show_value(U_bat, G_HIGH_SCALE_MIN, G_HIGH_SCALE_MAX);
            }
        } else {
            show_value(I_bat, G_HIGH_SCALE_MIN, G_HIGH_SCALE_MAX);
        }
    }
}

void second_tick() {
    // change mode max. once per second - protect the relay in weird situations

    if (U_bat < U_MIN || U_bat > U_MAX || I_bat > I_MAX) {
        mode = error;
        RELAY_LAT = 0;
        Out_duty = 0;
    }

    switch (mode) {

        case error:
            if (U_bat > U_MIN && U_bat < U_MAX) {
                RELAY_LAT = 1;
                mode = pre_charge;
            }
            break;

        case pre_charge:
            if (U_bat > U_DISCHARGED)
                mode = bulk_charge;
            break;

        case bulk_charge:
            if (U_bat > U_CHARGED)
                mode = absorb_charge;
            break;

        case absorb_charge:
            if (I_bat < I_MAINTAIN)
                mode = maintain_charge;
            break;
    }
}

void TMR1_CallBack(void) {
    static char duty;

    if (duty > 0)
        duty--;
    else {
        duty = 9;
        second_tick();
    }

    switch (mode) {

        case error:
            RED_LED_LAT = duty & 1;
            YELLOW_LED_LAT = 0;
            GREEN_LED_LAT = 0;
            break;

        case pre_charge:
            RED_LED_LAT = 0;
            YELLOW_LED_LAT = duty == 0;
            GREEN_LED_LAT = 0;
            break;

        case bulk_charge:
            RED_LED_LAT = 0;
            YELLOW_LED_LAT = duty < 5;
            GREEN_LED_LAT = 0;
            break;

        case absorb_charge:
            RED_LED_LAT = 0;
            YELLOW_LED_LAT = 0;
            GREEN_LED_LAT = duty < 5;
            break;

        case maintain_charge:
            RED_LED_LAT = 0;
            YELLOW_LED_LAT = 0;
            GREEN_LED_LAT = duty == 0;
            break;
    }
}

/**
 End of File
 */